#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from ..exceptions.ErroFormatacaoException import ErroFormatacaoException
from ..exceptions.InconsistenciaException import InconsistenciaException
from ..exceptions.FileNotFoundException import FileNotFoundException
from ..classes.DadosDeEntrada import DadosDeEntrada
from ..classes.Docente import Docente
from ..classes.Veiculo import Veiculo
from ..classes.Regra import Regra
from ..classes.Qualificacao import Qualificacao
from ..classes.Publicacao import Publicacao
from ..classes.PublicacaoConferencia import PublicacaoConferencia
from ..classes.PublicacaoPeriodico import PublicacaoPeriodico
# O arquivo ../exceptions/InconsistenciaException.py é um módulo de nome igual ao nome do arquivo
# A linha acima foi utilizada para importar apenas a classe InconsistenciaException desse módulo.
import locale
from datetime import datetime
import io

locale.setlocale(locale.LC_ALL, 'pt_BR.utf8') 

#Função que checa se o número é válido na string, e lança uma exceção caso não é.
def parseDouble(string):
    num = None
    try:
        num = locale.atof(string)
    except ValueError:
        raise ErroFormatacaoException()
    return num

class Leitura():
    #Classe responsável por ler os arquivos (planilhas) cujos nomes são definidos na construção de
    #objetos desta classe. Atributos  para lidar com local e formato de data são definidos como
    #constantes, pois não podem ser alteradas.

    def __init__(self, catQualis, nomeArqDoc, nomeArqVei, nomeArqPub, nomeArqQua, nomeArqReg):
        self.categoriasQualis = catQualis
        self.nomeArquivoQualis = nomeArqQua
        self.nomeArquivoRegras = nomeArqReg
        self.nomeArquivoDocentes = nomeArqDoc
        self.nomeArquivoPublicacoes = nomeArqPub
        self.nomeArquivoVeiculos = nomeArqVei

    #Função que lê a planilha de docentes cujo nome foi passado no construtor da classe Leitura;
    def __leDocentes(self, dadosEntrada):
        try:
            with io.open(self.nomeArquivoDocentes,"r",encoding="utf8") as arqEntrada:
                line = arqEntrada.readline()
                #Confere se a primeira linha do arquivo está correta. Se não estiver lança uma exceção.
                if line.strip() != u"Código;Nome;Data Nascimento;Data Ingresso;Coordenador?":
                    raise ErroFormatacaoException()
                for line in arqEntrada:
                    line = line.strip()
                    #Se a linha é vazia, a ignora
                    if (line == ""):
                        continue
                    #No caso da planilha de docentes, a String lida de uma linha deve ser dividida em pelo menos 4 campos
                    #contendo os dados de cada Docente. A divisão da String é feita através da funcao cpp_util::Tokenizer,
                    #e o vetor resultante tem os campos na seguinte ordem: Código;Nome;Data Nascimento;Data Ingresso;Coordenador?
                    #Onde o campo "Coordenador?" é opcional (apenas o coordenador possui este campo não vazio).

                    tokens = line.split(";");
                    if len(tokens) < 4:
                        raise ErroFormatacaoException()
                    #Elimina espaços no inicio e no fim da linha
                    tokens = map(unicode.strip,tokens) #aplica o método string.strip() a todos os elementos da lista tokens
                    #Os campos são convertidos de String para as respectivas classes
                    #Caso ocorra algum erro de formatação, uma exceção é gerada (e capturada na main)
                    codigo = None
                    try:
                        codigo = int(tokens[0])
                    except ValueError:
                        raise ErroFormatacaoException()
                    nome = tokens[1]
                    # A função strptime converte uma string (segundo um formato especificado)
                    # em uma estrutura do tipo "struct tm". Essa estrutura tem campos referentes
                    # a uma data (dia, mes, ano, etc...). 
                    dataNasc = None
                    dataIngresso = None
                    try:
                        dataNasc = datetime.strptime(tokens[2],"%d/%m/%Y",)
                        dataIngresso = datetime.strptime(tokens[3],"%d/%m/%Y")
                    except ValueError:
                        raise ErroFormatacaoException()
                    ehCoordenador = False
                    #Se vetor resultante do "corte" da string lida do arquivo tiver mais de 4 campos,
                    #Quer dizer que o docente é, possivelmente, coordenador.
                    if (len(tokens) > 4):
                        if (tokens[4] == "X"):
                            ehCoordenador = True
                    if dadosEntrada.codigoDocenteExiste(codigo):
                        raise InconsistenciaException("Código repetido para docente {}.".format(codigo))

                    #Cria um novo objeto do tipo docente e adiciona referências a este objeto nos dois mapas
                    #da classe DadosDeEntrada, ordenados pelo Código e pelo Nome do Docente;
                    #Essa inclusão do docente é feita através de um método da classe DadosDeEntrada;
                    docente = Docente(codigo,nome,dataNasc,dataIngresso,ehCoordenador)
                    dadosEntrada.adicionaDocente(docente)
        except IOError:
            raise FileNotFoundException()

    #Função que lê a planilha de veiculos cujo nome foi passado no construtor da classe Leitura;
    def __leVeiculos(self, dadosEntrada):
        try:
            with io.open(self.nomeArquivoVeiculos,"r",encoding="utf8") as arqEntrada:
                #Confere se a primeira linha do arquivo está correta. Se não estiver lança uma exceção.
                line = arqEntrada.readline()
                if line.strip() != u"Sigla;Nome;Tipo;Impacto;ISSN":
                    raise ErroFormatacaoException()
                for line in arqEntrada:
                    line = line.strip()
                    if (line == ""):
                        continue
                    #No caso da planilha de veiculos, a String lida de uma linha deve ser dividida em pelo menos 4 campos
                    #contendo os dados de cada Veiculo. A divisão da String é feita através da classe cpp_util::Tokenizer,
                    #e o vetor resultante tem os campos na seguinte ordem: Sigla;Nome;Tipo;Impacto;ISSN
                    #Onde o campo "ISSN" existe apenas em veículos do tipo "P", ou seja, periódicos.
                    tokens = line.split(";")
                    if len(tokens) < 4:
                        raise ErroFormatacaoException()
                    #Elimina espaços no inicio e no fim da linha
                    tokens = map(unicode.strip,tokens) #aplica o método string.strip() a todos os elementos da lista tokens
                    #Os campos são convertidos de String para as respectivas classes
                    #Caso ocorra algum erro de formatação, uma exceção é gerada (e capturada na main)
                    sigla = tokens[0]
                    nome = tokens[1]
                    tipo = tokens[2]
                    issn = None
                    # Usa a função parseDoubleWrap(string,loc) para converter string em double:
                    # Ela checa se a string é um número válido e chama a função cpp_util::parseDouble(s,l),
                    # Se a string não for válida, lança uma exceção do tipo ErroFormatacaoException();
                    fatorImpacto = 0.0
                    fatorImpacto = parseDouble(tokens[3]) #função lança exceção se falhar (capturada na main)
                    veiculo = None
                    if tipo == "C":
                        issn = ""
                    elif tipo == "P":
                        #Se for um veiculo do tipo "P", ele deve ter um ISSN
                        #Se não tiver, lança exceção de erro de formatação
                        if len(tokens) > 4:
                            issn = tokens[4]
                        else:
                            raise ErroFormatacaoException()
                    else:
                        raise InconsistenciaException(u"Tipo de veículo desconhecido para veículo {}: {}.".format(sigla,tipo))

                    #Cria um novo objeto do tipo veiculo 
                    veiculo = Veiculo(sigla,nome,tipo[0],fatorImpacto,issn)

                    #Caso sigla do veículo já exista, lança uma exceção E deleta o veículo criado acima
                    if dadosEntrada.siglaVeiculoExiste(veiculo.sigla):
                        raise InconsistenciaException(u"Código repetido para veículo: {}.".format(veiculo.sigla))

                    #adiciona referências a este objeto no mapa da classe DadosDeEntrada, ordenados pela Sigla do veículo
                    #Essa inclusão do veiculo é feita através de um método da classe DadosDeEntrada;
                    dadosEntrada.adicionaVeiculo(veiculo)
 
        except IOError:
            raise FileNotFoundException()

    #Função que lê a planilha de publicacoes cujo nome foi passado no construtor da classe Leitura;
    def __lePublicacoes(self, dadosEntrada):
        try:
            with io.open(self.nomeArquivoPublicacoes,"r",encoding="utf8") as arqEntrada:
                #Confere se a primeira linha do arquivo está correta. Se não estiver lança uma exceção.
                line = arqEntrada.readline()
                if line.strip() != u"Ano;Veículo;Título;Autores;Número;Volume;Local;Página Inicial;Página Final":
                    raise ErroFormatacaoException()
                #Enquanto tiver novas linhas...
                for line in arqEntrada:
                    #Se a linha é vazia, a ignora
                    if line == "":
                        continue
                    #No caso da planilha de publicacoes, a String lida de uma linha deve ser dividida em exatos 9 campos
                    #contendo os dados de cada publicacao. A divisão da String é feita através da funcao String.split(";"),
                    #e o vetor resultante tem os campos na seguinte ordem:
                    #Ano;Veículo;Título;Autores;Número;Volume;Local;Página Inicial;Página Final
                    #Onde o campo Volume existe apenas em publicações em periódicos e o campo Local só existe em
                    #publicações do tipo conferência. Embora as linhas desta planilha tenha campos opcionais, o primeiro e
                    #o último (ano e página final) são obrigatórios. Assim, a divisão da String sempre possuirá 9 campos.
                    tokens = line.split(";")
                    if len(tokens) < 9:
                        raise ErroFormatacaoException()
                    #Elimina espaços no inicio e no fim da linha
                    tokens = map(unicode.strip,tokens) #aplica o método string.strip() a todos os elementos da lista tokens
                    volume = None
                    numero = None
                    pagInicial = None
                    pagFinal = None
                    ano = None
                    siglaVeiculo = tokens[1]
                    titulo = tokens[2]
                    # O campo "autores" (4º campo ou índice 3 dos tokens) de uma publicação é
                    # uma lista de códigos (BigIntegers). A string é quebrada entre códigos
                    # (separados por vírgula), que são convertidos para BigInteger
                    listaAutores = tokens[3].split(",")
                    codigoAutores = []
                    local = tokens[6]
                    try:
                        #int() lança uma exceção "ValueError" se não conseguir interpretar
                        #o número, que se ocorrer, será capturada e "trocada" por uma exceção
                        #do tipo ErroFormatacaoException
                        numero = int(tokens[4])
                        if tokens[5] != "": #Volume só existe em publicacões de periódicos
                            volume = int(tokens[5])
                        pagInicial = int(tokens[7])
                        pagFinal = int(tokens[8])
                        ano = int(tokens[0]);
                        for codAutor in listaAutores:
                            #Checa se a String é um BigInt, caso contrário, 'converte' a exceção ValueError 
                            # em uma exceção ErroFormatacaoException()
                            codigoAutores.append(int(codAutor.strip()))
                    except ValueError:
                        raise ErroFormatacaoException()

                    publicacao = None
                    listaDocentesPublicacao = []

                    #Confere se o docente está cadastrado (consultando o mapa de docentes cuja chave é dada pelo
                    #do docente). Em caso negativo, lança uma exceção. Em caso afirmativo, adiciona o docente a uma
                    #lista encadeada de docentes da publicação. O docente é recuperado através do seu código.
                    for cod in codigoAutores:
                        if dadosEntrada.codigoDocenteExiste(cod):
                            listaDocentesPublicacao.append(dadosEntrada.getDocenteCodigo(cod))
                        else:
                            raise InconsistenciaException(u"Código de docente não definido usado na publicação \"{}\": {}.".format(titulo,cod))

                    #Confere se a sigla do veículo existe. Em caso negativo, lança uma exceção.

                    if not dadosEntrada.siglaVeiculoExiste(siglaVeiculo):
                        raise InconsistenciaException(u"Sigla de veículo não definida usada na publicação \"{}\": {}.".format(titulo,siglaVeiculo))

                    #Confere o tipo da publicação (classe abstrata) e cria um objeto da respectiva subclasse
                    if (local == ""): # Local só existe em conferencias
                        publicacao = PublicacaoPeriodico(ano,numero,pagInicial,pagFinal,titulo,listaDocentesPublicacao,volume)
                    else:
                        publicacao = PublicacaoConferencia(ano,numero,pagInicial,pagFinal,titulo,listaDocentesPublicacao,local);

                    #Acha o veiculo de acordo com siglaVeiculo
                    veiculo = dadosEntrada.getVeiculo(siglaVeiculo)

                    #Adiciona a publicação na lista de publicações do veículo
                    veiculo.adicionaPublicacao(publicacao)
                    #Adiciona uma referência do veículo encontrado no objeto da publicação.
                    publicacao.setVeiculo(veiculo)
                    #Adiciona a publicação na lista encadeada de publicaçoẽs do objeto da classe DadosDeEntrada.
                    dadosEntrada.adicionaPublicacao(publicacao)
                    #Adiciona a publicação na lista de publicação de todos os docentes autores desta publicação
                    for docente in listaDocentesPublicacao:
                        docente.adicionaPublicacao(publicacao)

        except IOError:
            raise FileNotFoundException()

    #Função que lê a planilha de qualis cujo nome foi passado no construtor da classe Leitura;
    def __leQualis(self, dadosEntrada):
        try:
            with io.open(self.nomeArquivoQualis,"r",encoding="utf8") as arqEntrada:
                line = arqEntrada.readline()
                #Confere se a primeira linha do arquivo está correta. Se não estiver lança uma exceção.
                if line.strip() != u"Ano;Veículo;Qualis":
                    raise ErroFormatacaoException()
                for line in arqEntrada:
                    line = line.strip()
                    #Se a linha é vazia, a ignora
                    if (line == ""):
                        continue
                    #No caso da planilha de qualificações, a String lida de uma linha deve ser dividida em exatos 3 campos
                    #contendo os dados de cada qualificação. A divisão da String é feita através da funcao String.split(";"),
                    #e o vetor resultante tem os campos na seguinte ordem: "Ano;Veículo;Qualis"
                    tokens = line.split(";")
                    if len(tokens) < 3:
                        raise ErroFormatacaoException()
                    #Elimina espaços no início e no fim da string.
                    tokens = map(unicode.strip,tokens) #aplica o método unicode.strip() a todos os elementos da lista tokens
                    #Os campos são convertidos de String para as respectivas classes
                    #Caso ocorra algum erro de formatação, uma exceção é gerada (e capturada na main)
                    siglaVeiculo = tokens[1]
                    qualificacao = tokens[2]
                    ano = None
                    try:
                        #Lança exceção ValueError se não conseguir converter
                        #Ela é "convertida" para ErroFormatacaoException
                        ano = int(tokens[0])
                    except ValueError:
                        raise ErroFormatacaoException();

                    #Procura no vetor de categorias se a categoria lida existe
                    #Confere se o qualis está cadastrado no sistema. Em caso negativo, lança uma exceção.
                    if qualificacao not in self.categoriasQualis:
                        raise InconsistenciaException(u"Qualis desconhecido para qualificação do veículo {} no ano {}: {}.".format(siglaVeiculo,ano,qualificacao))

                    #Confere se o veículo existe. Em caso negativo, lança uma exceção.
                    if not dadosEntrada.siglaVeiculoExiste(siglaVeiculo):
                        raise InconsistenciaException(u"Sigla de veículo não definida usada na qualificação do ano \"{}\": {}.".format(ano,siglaVeiculo))

                    #Acha o veículo correspondente da qualificação
                    veiculo = dadosEntrada.getVeiculo(siglaVeiculo)
                    #Cria um objeto da qualificação
                    qualif = Qualificacao(ano,qualificacao,veiculo)
                    #Adiciona a nova qualificação a lista de qualificações do veículo em questão.
                    veiculo.adicionaQualificacao(qualif)
        except IOError:
            raise FileNotFoundException()

    #Função que lê a planilha de publicacoes cujo nome foi passado no construtor da classe Leitura;
    def __leRegras(self, dadosEntrada):
        try:
            with io.open(self.nomeArquivoRegras,"r",encoding="utf8") as arqEntrada:
                line = arqEntrada.readline()
                #Confere se a primeira linha do arquivo está correta. Se não estiver lança uma exceção.
                if line.strip() != u"Início Vigência;Fim Vigência;Qualis;Pontos;Multiplicador;Anos;Mínimo Pontos":
                    raise ErroFormatacaoException()
                for line in arqEntrada:
                    line = line.strip()
                    #Se a linha é vazia, a ignora
                    if (line == ""):
                        continue
                    #No caso da planilha de regra, a String lida de uma linha deve ser dividida em exatos 7 campos
                    #contendo os dados de cada regra. A divisão da String é feita através da funcao String.split(";"),
                    #e o vetor resultante tem os campos na seguinte ordem:
                    #Início Vigência;Fim Vigência;Qualis;Pontos;Multiplicador;Anos;Mínimo Pontos
                    tokens = line.split(";")
                    if len(tokens) < 7:
                        raise ErroFormatacaoException()
                    #Elimina espaços no inicio e no fim da linha
                    tokens = map(unicode.strip,tokens) #aplica o método string.strip() a todos os elementos da lista tokens
                    #Os campos são convertidos de String para as respectivas classes
                    #Caso ocorra algum erro de formatação, uma exceção é gerada (e capturada na main)

                    dataInicio = None
                    dataFinal = None
                    qualisList = tokens[2].split(",")
                    pontosList = tokens[3].split(",")
                    multiplicador = 0.0
                    anos = 0
                    minimoPontos = 0.0
                    # A função strptime converte uma string (segundo um formato especificado)
                    # em uma estrutura do tipo "struct tm". Essa estrutura tem campos referentes
                    # a uma data (dia, mes, ano, etc...). 
                    try:
                        dataInicio = datetime.strptime(tokens[0],"%d/%m/%Y",)
                        dataFinal = datetime.strptime(tokens[1],"%d/%m/%Y")
                    except ValueError:
                        raise ErroFormatacaoException()

                    listaPontuacao = [0]*len(self.categoriasQualis) #Cria uma lista de zeros do tamanho da lista categoriasQualis

                    #O formato da lista de qualificações é, por exemplo; A1;B1;C com pontuações 10;5;0
                    #Isto significa que todas as qualificações entre A1 e B1 (excluindo B1)  valem 10 pontos;
                    #as entre B1 e C (excluindo C) valem 5 pontos; e todas as qualificações a partir de C valem 0;

                    #Para cada i-ésimo elemento da lista na forma A1,B1,C (exceto o último), percorre a lista de qualificações até
                    #encontrar a qualificação anterior a presente no elemento (i+1) desta lista, associando a i-ésima pontuação
                    #a estas qualificações. Caso seja o último da lista de qualis na planilha de regras, percorre o restante
                    #da lista associando a pontuação correspondente a elas.

                    #A existência de cada qualis é conferida. Em caso negativo, uma exceção é lançada.
                    #Esse for pega apenas os N-1 primeiros elementos da lista 
                    for index in xrange(0,len(qualisList)-1):
                        if qualisList[index] not in self.categoriasQualis:
                            #tokens[2] eh a string convertida para dataInicio
                            raise InconsistenciaException(u"Qualis desconhecido para regras de {}: {}".format(tokens[2],qualisList[index]))
                        try:
                            # Percorre a lista preenchendo todas as posições com qualis a partir do primeiro item
                            # até que ele encontre (preenche até uma posição antes) o indice do próximo (que já tem outra pontuação)
                            for index2 in xrange(self.categoriasQualis.index(qualisList[index]),self.categoriasQualis.index(qualisList[index+1])):
                                listaPontuacao[index2] = int(pontosList[index])
                        except ValueError:
                            raise ErroFormatacaoException()

                    #Se estiver no final da lista, as posições restantes devem ser preenchidas. Para isso,
                    # é necessário percorrer o vetor a partir da posição correspondente do último item da lista até o final deste:
                    try:
                        indexUltimoQualis = self.categoriasQualis.index(qualisList[-1])
                    except ValueError:
                        #tokens[2] eh a string convertida para dataInicio
                        raise InconsistenciaException(u"Qualis desconhecido para regras de {}: {}".format(tokens[2],qualisList[index]))

                    # substitui a parte do vetor citada acima por uma lista de tamanho correspondente onde os 
                    # valores são iguais a última pontuação fornecida
                    try:
                        listaPontuacao[indexUltimoQualis:] = [int(pontosList[-1])]*(len(listaPontuacao) - indexUltimoQualis)
                    except ValueError:
                        raise ErroFormatacaoException()

                    multiplicador = parseDouble(tokens[4]) #As funções lançam exceções se algo ocorrer errado
                    minimoPontos = parseDouble(tokens[6])  #exceção ErroFormatacaoException()

                    try:
                        anos = int(tokens[5])
                    except ValueError:
                        raise ErroFormatacaoException()

                    #Cria um novo objeto do tipo Regra.
                    regra = Regra(dataInicio,dataFinal,multiplicador,anos,minimoPontos)

                    #Adiciona as pontuações em um mapa (atributo da classe Regra) que tem como chave uma
                    #String (a qualificação) e sua pontuação como valor.
                    for index in xrange(len(listaPontuacao)):
                        regra.adicionaPontuacao(self.categoriasQualis[index],listaPontuacao[index])

                    dadosEntrada.adicionaRegra(regra);

        except IOError:
            raise FileNotFoundException()

    #Função publica para ler os dados de entrada a partir dos arquivos
    def leArquivos(self):
        dadosEntrada = DadosDeEntrada()
        self.__leDocentes(dadosEntrada)
        self.__leVeiculos(dadosEntrada)
        self.__lePublicacoes(dadosEntrada)
        self.__leQualis(dadosEntrada)
        self.__leRegras(dadosEntrada)
        return dadosEntrada
